#include <process.h>
#include <stdlib.h>
#include <assert.h>
#include <malloc.h>
#include <stdarg.h>
#include <conio.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>
#include <dos.h>
#include <i86.h>
#include <mem.h>

#define IRQ0_FREQ 36
#define NUM_CLOUDS 4
#define NUM_TREES 4
#define SANITY_CHECKS

#define MOVE_SUN 1
#define MOVE_CLOUDS 2
#define MOVE_GROUNDMAP 4	/* dotyczy te drzew */

#pragma aux setvmode = "int 0x10", parm [ax]
#define sync_buffers() memcpy ((char far *) 0xA0000000L, vbuf, 64000)

/*
 * na pocztku: PAL_FADEIN
 * podczas trwania: PAL_SUSTAIN
 * po escape: PAL_FADEOUT
 * gdy PAL_DONE, wyjcie
 */

enum {PAL_FADEIN = 0, PAL_SUSTAIN, PAL_FADEOUT, PAL_DONE};

void setvmode (unsigned short int);
char far *vbuf;
char pal_state = PAL_FADEIN;
char move_mask = 0;

char pal[768], keytab[128], groundmap[320];
signed short int *sincos;

void (__interrupt __far *old_irq0) (void);
void (__interrupt __far *old_irq1) (void);

struct sun_typ {
	char *pix;	/* 64x64 */
	signed short x, y, base_x;
	short angle;
	short hop_var;
	char old_pal_state;
	signed short saved_base;
	int draw_phase;
} sun;

struct cloud_typ {
	char *pix;	/* 64x32 */
	signed short x, y;
	int dx[8], dy[8];
	int draw_phase;
	int cur_ang;
} cloud[NUM_CLOUDS];

struct tree_typ {
	char *pix;		/* 64x128 */
	signed short x, y;
	int cx[8], cy[8];
	int dx[8], dy[8];
	int draw_phase;
	int cur_ang;
} tree[NUM_TREES];

void do_pal (void)
{
	static char mypal[768];
	static char first_time = 1;
	int i;
	char done = 1;

	if (pal_state == PAL_SUSTAIN || pal_state == PAL_DONE)
		return;

	if (first_time) {
		first_time = 0;
		if (pal_state == PAL_FADEIN)
			bzero (mypal, sizeof(mypal));
		else
			memcpy (mypal, pal, sizeof(mypal));
	}

	if (pal_state == PAL_FADEIN) {
		for (i = 0; i < 768; i++)
			if (mypal[i] < pal[i]) {
				mypal[i]++;
				done = 0;
			}
	} else
		for (i = 0; i < 768; i++)
			if (mypal[i]) {
				mypal[i]--;
				done = 0;
			}

	if (done) {
		if (pal_state == PAL_FADEIN)
			pal_state = PAL_SUSTAIN;
		else
			pal_state = PAL_DONE;
		return;
	}

	outp (0x3C8, 0x00);
	for (i = 0; i < 768; i++)
		outp (0x3C9, mypal[i]);
}

void phase_sun (void);
void phase_clouds (void);
void phase_trees (void);

void move_sun (void)
{
	phase_sun();
	if (pal_state == PAL_FADEOUT) {
		if (sun.old_pal_state != PAL_FADEOUT) {
			sun.old_pal_state = PAL_FADEOUT;
			sun.saved_base = sun.base_x;
		}

		if (sun.base_x > -64)
			sun.base_x -= (sun.saved_base > (160 - 32)) ? 1 : 2;
	} else
		if (sun.base_x > (160 - 32))
			sun.base_x--;

	if (sun.hop_var) {
		sun.x = sun.base_x + (sincos[sun.angle ^ 1024] * (sun.hop_var >> 5) >> 10);
		sun.y = sincos[sun.angle] * (sun.hop_var >> 3) >> 10;
		sun.angle += 16;
		sun.angle &= 1023;
		if (sun.base_x == (160 - 32)) {
			if ((sun.hop_var >> 3) > 3)
				sun.hop_var--;
		} else if (sun.base_x < (160 - 32))
			if ((sun.hop_var >> 3) < 16)
				sun.hop_var++;
	}
}

void move_cloud (int i)
{
	struct cloud_typ *ptr = &cloud[i];
	static char div = 0;

	if (div >= i)
		if (ptr->x > -64)
			ptr->x -= (rand() & 1) + 1;
		else
			ptr->x = 320;

	if (!i) {
		div++;
		if (div == NUM_CLOUDS)
			div = 0;
	}
}

void move_clouds (void)
{
	int i;

	phase_clouds();
	for (i = 0; i < NUM_CLOUDS; i++)
		move_cloud (i);
}

void move_tree (int i)
{
	struct tree_typ *ptr = &tree[i];

	if (ptr->x > -64) {
		ptr->x--;
		if (ptr->x == 319)
			if (tree->draw_phase >= (128 - 32 + 128 + 128 + 128))
//				ptr->y = groundmap[319] - 128 + 6 + (rand() & 0x5F);	/* evil hack */
				ptr->y = groundmap[319] - 128 + 6 + 0x5F;	/* evil hack */
			else
				ptr->y = groundmap[319] - 128 + 6 + (rand() & 0x2F);
	} else
		ptr->x = (rand() % 320) + 320;

	if (ptr->y > 140 && tree->draw_phase >= (128 - 32 + 128 + 128 + 128))
		ptr->y--;
}

void move_groundmap (void)
{
	int i, x, y;
	signed int rndv;

	for (x = 1; x < 320; x++)
		groundmap[x - 1] = groundmap[x];

	y = groundmap[319];
	rndv = (rand() % 3) - 1;
	if ((y + rndv) > (199 - 16) || (y + rndv) < (199 - 64))
		y -= rndv;
	else
		y += rndv;

	groundmap[319] = y;

	phase_trees();

	for (i = 0; i < NUM_TREES; i++)
		move_tree (i);
}

void do_move (void)
{
	if (!(move_mask & MOVE_SUN))
		move_sun();
	if (!(move_mask & MOVE_CLOUDS))
		move_clouds();
	if (!(move_mask & MOVE_GROUNDMAP))
		move_groundmap();
}

void irq0 (void)
{
	static char div = 0;
	if (div++ >= 1) {
		div = 0;
		do_pal();
	}
	do_move();
}

void __interrupt __far new_irq0 (void)
{
	static unsigned short int local_ticks = 0;

	irq0();

	if (++local_ticks >= (65535L * IRQ0_FREQ / 1193180L)) {
		local_ticks = 0;
		_chain_intr (old_irq0);
	} else _asm {
		mov al, 20h
		out 20h, al
	};
}

void __interrupt __far new_irq1 (void)
{
	unsigned char ch;

	ch = inp(0x60);
	_asm {
		in	al, 61h
		or	al, 82h
		out	61h, al
		and	al, 7Fh
		out	61h, al
	};

	keytab[ch & 0x7F] = (ch >> 7) ^ 1;
	if (ch == 0x01)
		pal_state = PAL_FADEOUT;
	outp (0x20, 0x20);
}

void loadpal (void)
{
	FILE *fp = fopen("palette.pal", "rb");
	assert (fp);
	fread (pal, 768, 1, fp);
	fclose (fp);
}

void make_sun (void)
{
	/* zainicjowanie struktury */

	sun.pix = (char *) malloc(4096);
	bzero (sun.pix, 4096);
	sun.x = sun.base_x = 320;
	sun.y = 0;
	sun.angle = 0;
	sun.hop_var = 16 << 3;
	sun.old_pal_state = PAL_FADEIN;
	sun.saved_base = 0;
	sun.draw_phase = 0;
}

void phase_sun (void)
{
	int x, y;
	int a;
	static int cur_ang = 0;
	unsigned int ofs;

	/* najpierw kko */

	if (sun.draw_phase < 64) {
		for (a = cur_ang; a < (cur_ang + 16); a++) {
			x = sincos[a ^ 1024] * 16 / 1024 + 32;
			y = sincos[a] * 16 / 1024 + 32;
			sun.pix[(y << 6) + x] = 1;
		}
		cur_ang += 16;
	} else if (sun.draw_phase < (64 + (47 - 18))) {
		char fill = 0, px, old_px = 0;

		for (x = 17; x < 47; x++) {
			ofs = (((sun.draw_phase - 64) + 18) << 6) ^ x;
#ifdef	SANITY_CHECKS
			assert (ofs < 4096);
#endif
			px = sun.pix[ofs];

			if (px && old_px)
				continue;

			if (px)
				fill ^= 1;

			if (fill && !px)
				sun.pix[ofs] = 6;

			old_px = px;
		}
	} else if (sun.draw_phase < (64 + (47 - 18) + 64)) {
		ofs = (32 << 6) + (sun.draw_phase - (64 + (47 - 18)));
#ifdef	SANITY_CHECKS
		assert (ofs < 4096);
#endif
		sun.pix[ofs] = sun.pix[ofs] ? sun.pix[ofs] : 1;
		ofs = ((sun.draw_phase - (64 + (47 - 18))) << 6) + 32;
#ifdef	SANITY_CHECKS
		assert (ofs < 4096);
#endif
		sun.pix[ofs] = sun.pix[ofs] ? sun.pix[ofs] : 1;
	} else if (sun.draw_phase < (64 + (47 - 18) + 64 + (56 - 8))) {
		ofs = ((sun.draw_phase - (64 + (47 - 18) + 64 - 8)) << 6) ^ (sun.draw_phase - (64 + (47 - 18) + 64 - 8));
#ifdef	SANITY_CHECKS
		assert (ofs < 4096);
#endif
		sun.pix[ofs] = sun.pix[ofs] ? sun.pix[ofs] : 1;
		ofs = ((sun.draw_phase - (64 + (47 - 18) + 64 - 8)) << 6) ^ ((sun.draw_phase - (64 + (47 - 18) + 64 - 8)) ^ 63);
#ifdef	SANITY_CHECKS
		assert (ofs < 4096);
#endif
		sun.pix[ofs] = sun.pix[ofs] ? sun.pix[ofs] : 1;
	} else if (sun.draw_phase < (64 + (47 - 17) + 64 + (56 - 8)) + 1) {
		sun.pix[(26 << 6) + 26] = 3;
		sun.pix[(26 << 6) + 38] = 3;
	} else if (sun.draw_phase < (64 + (47 - 17) + 64 + (56 - 8)) + 1 + 1) {
		sun.pix[(38 << 6) + 26] = 4;
		sun.pix[(39 << 6) + 27] = 4;
		for (x = 28; x <= 36; x++)
			sun.pix[(40 << 6) + x] = 4;
		sun.pix[(39 << 6) + 37] = 4;
		sun.pix[(38 << 6) + 38] = 4;
	}

	if (sun.draw_phase < (64 + (47 - 17) + 64 + (56 - 8)) + 1 + 1 + 1)
		sun.draw_phase++;
}

void fill_cloud_border (char *pix, int x, int y)
{
	char *buf = pix + (y << 6) + x;

#ifdef	SANITY_CHECKS
	assert (((y << 6) ^ x) < 2048);
#endif

	*buf = *buf == 1 ? *buf : 2;
}

void fill_tree_border (char *pix, int x, int y)
{
	char *buf = pix + (y << 6) + x;

#ifdef	SANITY_CHECKS
	assert (((y << 6) ^ x) < 8192);
#endif

	*buf = !((*buf & 0x60) ^ 0x60) ? *buf : 0x01;
}

void make_cloud (struct cloud_typ *cloud)
{
	int i;

	cloud->pix = (char *) malloc(2048);
	bzero (cloud->pix, 2048);
	cloud->x = rand() % (320 + 64) - 64;
	cloud->y = (rand() & 15) + 32;
	cloud->draw_phase = 0;
	cloud->cur_ang = 0;
	for (i = 0; i < 8; i++) {
		cloud->dx[i] = rand() % (47 - 16) + 16;
		cloud->dy[i] = rand() % (23 - 8) + 8;
	}
}

void phase_cloud (struct cloud_typ *cloud)
{
	int x, y;
	int a, i;
	int first, last;
	int ofs;

	if (cloud->draw_phase < 64) {
		for (i = 0; i < 8; i++) {
			for (a = cloud->cur_ang; a < cloud->cur_ang + 16; a++) {
				x = ((sincos[a ^ 1024] * 16) / 1024) + cloud->dx[i];
				y = ((sincos[a] * 8) / 1024) + cloud->dy[i];
				ofs = (y << 6) ^ x;
#ifdef	SANITY_CHECKS
				assert (ofs < 2048);
#endif
				cloud->pix[ofs] = 1;
			}
		}
		cloud->cur_ang += 16;
	} else if (cloud->draw_phase < (64 + 32)) {
		ofs = (cloud->draw_phase - 64) << 6;
		first = -1;

		for (x = 0; x < 64; x++) {
#ifdef	SANITY_CHECKS
			assert ((ofs + x) < 2048);
#endif
			if (cloud->pix[ofs + x]) {
				if (first == -1)
					first = x;
				last = x;
			}
		}

		if (first != -1 && first != last) {
#ifdef	SANITY_CHECKS
			assert ((ofs + first) < 2048);
			assert ((ofs + last) < 2048);
#endif
			memset (cloud->pix + ofs + first, 1, last - first + 1);
		}
	} else if (cloud->draw_phase < (64 + 32 + 32)) {
		y = cloud->draw_phase - (64 + 32);
		for (x = 0; x < 64; x++) {
			if (cloud->pix[(y << 6) ^ x] != 1)
				continue;
			if (x > 0) {
				fill_cloud_border (cloud->pix, x - 1, y);
				if (y > 0)
					fill_cloud_border (cloud->pix, x - 1, y - 1);
				if (y < 31)
					fill_cloud_border (cloud->pix, x - 1, y + 1);
			}
			if (y > 0)
				fill_cloud_border (cloud->pix, x, y - 1);
			if (y < 31)
				fill_cloud_border (cloud->pix, x, y + 1);
			if (x < 63) {
				fill_cloud_border (cloud->pix, x + 1, y);
				if (y > 0)
					fill_cloud_border (cloud->pix, x + 1, y - 1);
				if (y < 31)
					fill_cloud_border (cloud->pix, x + 1, y + 1);
			}
		}
	}

	if (cloud->draw_phase < (64 + 32 + 32 + 32))
		cloud->draw_phase++;
}

void make_clouds (void)
{
	int i;

	for (i = 0; i < NUM_CLOUDS; i++)
		make_cloud (&cloud[i]);
}

void phase_clouds (void)
{
	int i;

	for (i = 0; i < NUM_CLOUDS; i++)
		phase_cloud (&cloud[i]);
}

void make_tree (struct tree_typ *tree)
{
	int i;
	static char ftime = 1;
	/* tree->pix: 64x128 */

	tree->pix = (char *) malloc(8192);
	if (ftime) {
		tree->x = rand() % 150 + 160;
		tree->y = groundmap[tree->x] - 128 + 6 + (rand() & 0x2F);
		ftime = 0;
	} else {
		tree->x = (rand() % 320) + 320;
		tree->y = 0;	/* l8r */
	}
	bzero (tree->pix, 8192);
	for (i = 0; i < 8; i++) {
		tree->dx[i] = sincos[((i * 1024) / 8) ^ 1024] * 16 / 1024 - 3 + (rand() & 7);
		tree->dy[i] = sincos[(i * 1024) / 8] * 16 / 1024 - 3 + (rand() & 7);
		tree->cx[i] = (rand() & 3) | 8;
		tree->cy[i] = (rand() & 3) | 8;
	}
	tree->draw_phase = tree->cur_ang = 0;
}

void phase_tree (struct tree_typ *tree)
{
	int x, y;
	int a, i;
	int ofs;
	signed int first, last;

	if (tree->draw_phase < (128 - 32)) {
		y = tree->draw_phase + 32;
#ifdef SANITY
		assert ((y << 6) + 28);
		assert ((y << 6) + 35);
#endif
		tree->pix[(y << 6) + 28] = tree->pix[(y << 6) + 35] = 0x50;
		for (x = 29; x < 35; x++)
			tree->pix[(y << 6) + x] = (rand() & 0x0F) + 0x50;
	} else if (tree->draw_phase < (128 - 32 + 128)) {
		for (i = 0; i < 8; i++)
			for (a = tree->cur_ang; a < tree->cur_ang + 8; a++) {
				x = (sincos[a ^ 1024] * tree->cx[i]) / 1024 + 31 + tree->dx[i];
				y = (sincos[a] * tree->cy[i]) / 1024 + 31 + tree->dy[i];
				ofs = (y << 6) + x;
#ifdef SANITY
				assert (ofs < 8192);
#endif
				tree->pix[ofs] = (rand() & 0x0F) + 0x60;
			}
			tree->cur_ang += 8;
	} else if (tree->draw_phase < (128 - 32 + 128 + 128)) {
		ofs = (tree->draw_phase - (128 - 32 + 128)) << 6;
		first = -1;

		for (x = 0; x < 64; x++) {
#ifdef	SANITY_CHECKS
			assert ((ofs + x) < 8192);
#endif
			if ((tree->pix[ofs + x] & 0x60) == 0x60) {
				if (first == -1)
					first = x;
				last = x;
			}
		}

		if (first != -1 && first != last) {
#ifdef	SANITY_CHECKS
			assert ((ofs + first) < 8192);
			assert ((ofs + last) < 8192);
#endif
			for (i = first; i <= last; i++)
				tree->pix[ofs + i] = (rand() & 0x0F) | 0x60;
		}
	} else if (tree->draw_phase < (128 - 32 + 128 + 128 + 128)) {
		y = tree->draw_phase - (128 - 32 + 128 + 128);
		for (x = 0; x < 64; x++) {
			if ((tree->pix[(y << 6) ^ x] & 0x60) != 0x60)
				continue;
			if (x > 0) {
				fill_tree_border (tree->pix, x - 1, y);
				if (y > 0)
					fill_tree_border (tree->pix, x - 1, y - 1);
				if (y < 127)
					fill_tree_border (tree->pix, x - 1, y + 1);
			}
			if (y > 0)
				fill_tree_border (tree->pix, x, y - 1);
			if (y < 127)
				fill_tree_border (tree->pix, x, y + 1);
			if (x < 63) {
				fill_tree_border (tree->pix, x + 1, y);
				if (y > 0)
					fill_tree_border (tree->pix, x + 1, y - 1);
				if (y < 127)
					fill_tree_border (tree->pix, x + 1, y + 1);
			}
		}
	}

	if (tree->draw_phase < (128 - 32 + 128 + 128 + 128))
		tree->draw_phase++;
}

void make_trees (void)
{
	int i;

	for (i = 0; i < NUM_TREES; i++)
		make_tree (&tree[i]);
}

void phase_trees (void)
{
	int i;

	for (i = 0; i < NUM_TREES; i++)
		phase_tree (&tree[i]);
}

void make_groundmap (void)
{
	int x, y = 199;
	signed int rndv;

	for (x = 0; x < 320; x++) {
		rndv = (rand() % 3) - 1;
		if ((y + rndv) > (199 - 16) || (y + rndv) < (199 - 64))
			y -= rndv;
		else
			y += rndv;

		if (y >= 200)
			y = 199;

		groundmap[x] = y;
	}
}

void initscreen (void)
{
	unsigned short i;

	make_sun();
	make_clouds();
	make_groundmap();
	make_trees();

	outp (0x3C6, 0xFF);
	outp (0x3C8, 0x00);
	for (i = 0; i < 768; i++)
		outp (0x3C9, 0x00);
}

void init (void)
{
	FILE *fp;
	int fq;

	sincos = (signed short int *) malloc(2048 * sizeof(signed short int));
	fp = fopen("tables.dat", "rb");
	assert (fp);
	fread (sincos, 2048, sizeof(signed short int), fp);
	fclose (fp);
	vbuf = (char far *) _fmalloc(64000);
	bzero (vbuf, 64000);
	bzero (keytab, 128);
	loadpal();
	setvmode (0x13);
	initscreen();
	old_irq0 = _dos_getvect(0x08);
	old_irq1 = _dos_getvect(0x09);
	fq = 1193180L / IRQ0_FREQ;
	_asm {
		cli
		mov al, 3Ch
		out 43h, al
		mov ax, fq
		xchg ah, al
		out 40h, al
		xchg ah, al
		out 40h, al
	}
	_dos_setvect (0x08, new_irq0);
	_dos_setvect (0x09, new_irq1);
	_asm sti;
}

void deinit (void)
{
	int i;

	_asm {
		cli
		mov al, 3Ch
		out 43h, al
		mov al, 0FFh
		out 40h, al
		out 40h, al
	}
	_dos_setvect (0x09, new_irq1);
	_dos_setvect (0x08, new_irq0);
	_asm sti;
	free (sincos);
	free (vbuf);
	free (sun.pix);
	for (i = 0; i < NUM_TREES; i++)
		free (tree[(NUM_TREES - 1) - i].pix);
	free (tree[0].pix);
	for (i = 0; i < NUM_CLOUDS; i++)
		free (cloud[(NUM_CLOUDS - 1) - i].pix);
}

void draw_sun (void)
{
	int x, y;
	int start_x, start_y, end_x, end_y;
	signed short sun_x, sun_y;

	move_mask |= MOVE_SUN;
	sun_x = sun.x;
	sun_y = sun.y;
	move_mask &= ~MOVE_SUN;

	if (sun_x <= -64 || sun_y <= -64 || sun_x >= 320 || sun_y >= 200)
		return;

	if (sun_x < 0)
		start_x = -sun_x;
	else
		start_x = 0;

	if (sun_y < 0)
		start_y = -sun_y;
	else
		start_y = 0;

	if (sun_x > (320 - 64))
		end_x = 64 - (sun_x - (320 - 64));
	else
		end_x = 64;

	if (sun_y > (200 - 64))
		end_y = 64 - (sun_y - (200 - 64));
	else
		end_y = 64;

	for (y = start_y; y < end_y; y++)
		for (x = start_x; x < end_x; x++)
			if (sun.pix[(y << 6) + x])
				vbuf[((sun_y + y) << 8) + ((sun_y + y) << 6) + sun_x + x] = sun.pix[(y << 6) + x];
}

void draw_cloud (struct cloud_typ *cloud, char color)
{
	int x, y;
	int start_x, start_y, end_x, end_y;
	signed short cl_x, cl_y;

	move_mask |= MOVE_CLOUDS;
	cl_x = cloud->x;
	cl_y = cloud->y;
	move_mask &= ~MOVE_CLOUDS;

	if (cl_x <= -64 || cl_y <= -32 || cl_x >= 320 || cl_y >= 200)
		return;

	if (cl_x < 0)
		start_x = -cl_x;
	else
		start_x = 0;

	if (cl_y < 0)
		start_y = -cl_y;
	else
		start_y = 0;

	if (cl_x > (320 - 64))
		end_x = 64 - (cl_x - (320 - 64));
	else
		end_x = 64;

	if (cl_y > (200 - 32))
		end_y = 32 - (cl_y - (200 - 32));
	else
		end_y = 32;

	for (y = start_y; y < end_y; y++)
		for (x = start_x; x < end_x; x++) {
			char px = cloud->pix[(y << 6) + x];
			if (px)
				vbuf[((cl_y + y) << 8) + ((cl_y + y) << 6) + cl_x + x] = (px == 2) ? color : color + 6;
		}
}

void draw_clouds (void)
{
	int i;

	for (i = 0; i < NUM_CLOUDS; i++)
		draw_cloud (&cloud[(NUM_CLOUDS - 1) - i], 16 + (((NUM_CLOUDS - 1) - i) << 2));
}

void draw_tree (struct tree_typ *tree, int tx, int ty)
{
	int x, y;
	int start_x, start_y, end_x, end_y;
	signed short cl_x, cl_y;

	cl_x = tx;
	cl_y = ty;

	if (cl_x <= -64 || cl_y <= -128 || cl_x >= 320 || cl_y >= 200)
		return;

	if (cl_x < 0)
		start_x = -cl_x;
	else
		start_x = 0;

	if (cl_y < 0)
		start_y = -cl_y;
	else
		start_y = 0;

	if (cl_x > (320 - 64))
		end_x = 64 - (cl_x - (320 - 64));
	else
		end_x = 64;

	if (cl_y > (200 - 128))
		end_y = 128 - (cl_y - (200 - 128));
	else
		end_y = 128;

	for (y = start_y; y < end_y; y++)
		for (x = start_x; x < end_x; x++) {
			char px = tree->pix[(y << 6) + x];
			if (px == 0x01)
				px = 0x60;
			if (px)
				vbuf[((cl_y + y) << 8) + ((cl_y + y) << 6) + cl_x + x] = px;
		}
}

void draw_trees (int *x, int *y)
{
	int i;

	for (i = 0; i < NUM_TREES; i++)
		draw_tree (&tree[(NUM_TREES - 1) - i], x[i], y[i]);
}

void draw_groundmap (void)
{
	char map[320];
	int x, y, yp;
	int i, xy[2][NUM_TREES];

	move_mask |= MOVE_GROUNDMAP;
	memcpy (map, groundmap, 320);
	for (i = 0; i < NUM_TREES; i++) {
		xy[0][i] = tree[i].x;
		xy[1][i] = tree[i].y;
	}
	move_mask &= ~MOVE_GROUNDMAP;

	draw_trees (xy[0], xy[1]);

	for (x = 0; x < 320; x++) {
		yp = map[x];
		for (y = yp; y < 200; y++)
			vbuf[(y << 8) + (y << 6) + x] = 2;
		vbuf[(yp << 8) + (yp << 6) + x] = 5;
	}
}

void make_view (void)
{
	int y;
	char far *vptr = vbuf;

	for (y = 0; y < 32; y++) {
		memset (vptr, y + 16, 1920);
		vptr += 1920;
	}
	memset (vptr, y + 16 - 1, 2560);
	draw_sun();
	draw_clouds();
	draw_groundmap();
}

int main (int argc, char **argv)
{
	init();
	while (pal_state != PAL_DONE) {
		make_view();
		sync_buffers();
	}
	deinit();
	return 0;
}
