/*
 * init_objects(): inicjalizacja obiektw, struktur danych, pamici, itp. wane 
 * eby struktury obsugiwane wspbienie przez move_objects() i draw_objects() 
 * byy alokowane statycznie, inaczej podczas kopiowania w draw_objects() zostan 
 * skopiowane tylko wskaniki do dynamicznie zaalokowanej pamici.
 *
 * deinit_objects(): zwolnienie zasobw zainicjowanych przez init_objects().
 *
 * move_objects(): zaktualizowanie struktur opisujcych obiekty.
 *
 * draw_objects(): narysowanie obiektw w vbuf. oprcz tego, jeli faza rysowania 
 * zostaa zaktualizowana przez move_objects(), funkcja draw_objects() aktualizuje 
 * obrazki (wywoujc dla kadego obiektu phase_object()) tyle razy, ile wynosi 
 * rnica midzy now i star faz. oprcz tego draw_objects() kopiuje struktury 
 * opisujce poszczeglne obiekty do lokalnej struktury przed operacjami na nich, 
 * wyczajc na czas tej operacji dziaanie move_objects() (w kadym momencie 
 * draw_objects() moe zosta przerwane przez move_objects(), a draw_objects() 
 * do odpowiedniego pokazania obiektw uywa danych aktualizowanych przez 
 * move_objects()).
 *
 * hack: phase (faza) jest faz rysowania obiektu, wykorzystywan na pocztku 
 * caej animacji do pynnego rysowania poszczeglnych obiektw, jednak obiekty 
 * s aktualizowane dopiero przed narysowaniem. procedura move_objects() nie 
 * rusza obrazkw - one s aktualizowane dopiero w draw_objects(), przed 
 * obciciem, zamaskowaniem przezroczystoci i narysowaniem w buforze ekranu. 
 * ograniczenia pamici i wspbienoci wykonywania, opisane wyej.
 *
 * przy tworzeniu nowego obiektu trzeba zdefiniowa:
 *
 *  * struktur danych opisujc wszystkie cechy obiektu + wpis w pos_type.
 *
 *  * wskanik na obraz obiektu w strukturze pix_type.
 *
 *  * init_object(): inicjalizacja obiektu + wpis w init_objects().
 *
 *  * deinit_object(): deinicjalizacja obiektu + wpis w deinit_objects().
 *
 *  * move_object(): przesuwanie obiektu + wpis w move_objects().
 *
 *  * phase_object (int phase): narysowanie fazy obiektu.
 *
 *  * draw_object (struct pos_type *pos): narysowanie obiektu o okrelonych 
 *    cechach, zawartych w strukturze pos_type, wskazywanej przez *pos.
 *
 */

#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <mem.h>

enum {
	ST_FADEIN = 0,
	ST_SUSTAIN,
	ST_FADEOUT,
	ST_DONE
};

extern char far *vbuf;
extern char *keytab;
extern signed short int *sincos;
extern char pstate;

static char can_move = 1;

#define NCLOUD 4
#define NTREE 8
#define MAX_PHASE 3000

struct sun_type {
	signed short bx, saved_bx, x, y;
	short hop, ang;
	char opstate;
};

struct cloud_type {
	signed short x, y;
	int dx[8], dy[8];
};

struct ground_type {
	char map[320];	/* musi by static */
	char shift;
};

struct tree_type {
	signed short x, y;
	signed short dx[8], dy[8];
};

struct mouse_type {
	signed short x, y;
	char phase;
};

struct pos_type {
	struct sun_type sun;
	struct cloud_type cloud[NCLOUD];
	struct ground_type ground;
	struct tree_type tree[NTREE];
	struct mouse_type mouse;
	int phase;
};

struct pix_type {
	char *sun;
	char *cloud[NCLOUD];
	char *ground;
	char *tree[NTREE];
	char *mouse[2];
};

struct obj_type {
	struct pix_type pix;
	struct pos_type pos;
} obj;

void draw_picture (char *pix, int rsx, int rsy, int px, int py)
{
	int x, y;
	int sx, sy, ex, ey;
	char sft, pixel;

	for (sft = 0; sft < 8; sft++)
		if ((1 << sft) == rsx)
			break;

	if (px <= -rsx || py <= -rsy || px >= 320 || py >= 200)
		return;

	if (px < 0)
		sx = -px;
	else
		sx = 0;

	if (py < 0)
		sy = -py;
	else
		sy = 0;

	if (px > (320 - rsx))
		ex = rsx - (px - (320 - rsx));
	else
		ex = rsx;

	if (py > (200 - rsy))
		ey = rsy - (py - (200 - rsy));
	else
		ey = rsy;

	for (y = sy; y < ey; y++)
		for (x = sx; x < ex; x++) {
			pixel = pix[(y << sft) + x];
			if (pixel)
				vbuf[((py + y) << 8) + ((py + y) << 6) + px + x] = pixel;
		}
}

char gcolor (char start, char end)
{
	static char c, cdir = 1;
	static char ostart = -1, oend;

	if (ostart != start || oend != end) {
		ostart = start;
		oend = end;
		c = start;
		cdir = 1;
	}

	if (rand() & 1)
		if (cdir)
			if (c < end)
				c++;
			else
				cdir ^= 1;
		else
			if (c > start)
				c--;
			else
				cdir ^= 1;

	return c;
}

void fill_line (char *line, int len, char start, char end, char fillall, char random)
{
	int i;
	int si = -1, ei;

	for (i = 0; i < len; i++) {
		if (!line[i])
			continue;

		if (random && line[i] >= 4 && line[i] <= 35)
			continue;

		if (si == -1)
			si = i;

		ei = i;
	}

	if (si == -1)
		return;

	for (i = si; i <= ei; i++)
		if (fillall || !line[i])
			if (random)
				line[i] = (rand() % ((end - start) + 1)) + start;
			else
				line[i] = gcolor(start, end);
}

void init_sun (void)
{
	obj.pix.sun = (char *) malloc(64 * 64);
	bzero (obj.pix.sun, 64 * 64);
	obj.pos.sun.x = 320;
	obj.pos.sun.y = 0;
	obj.pos.sun.bx = 320;
	obj.pos.sun.opstate = pstate;
	obj.pos.sun.hop = 16 << 3;
	obj.pos.sun.ang = 0;
}

void init_cloud (void)
{
	int i, n;

	for (i = 0; i < NCLOUD; i++) {
		obj.pix.cloud[i] = (char *) malloc(64 * 32);
		bzero (obj.pix.cloud[i], 64 * 32);
		obj.pos.cloud[i].x = (rand() % (320 + 64)) - 64;
		obj.pos.cloud[i].y = rand() % 48;
		for (n = 0; n < 8; n++) {
			obj.pos.cloud[i].dx[n] = (rand() % ((64 - 17) - 17)) + 17;
			obj.pos.cloud[i].dy[n] = (rand() % ((32 - 9) - 9)) + 9;
		}
	}
}

void init_groundmap (void)
{
	int x;
	int y = (rand() & 31) + 150, d;
	int i;

	obj.pix.ground = (char *) malloc(16 * 16);
	for (i = 0; i < 256; i++)
		obj.pix.ground[i] = (rand() % ((103 - 71) - 1)) + 71;

	for (x = 0; x < 320; x++) {
		obj.pos.ground.map[x] = y;
		d = (rand() % 3) - 1;
		if ((y + d) > (150 + 32) || (y + d) < 150)
			y -= d;
		else
			y += d;
	}

	obj.pos.ground.shift = 0;
}

void init_tree (void)
{
	int i, n;

	for (i = 0; i < NTREE; i++) {
		obj.pix.tree[i] = (char *) malloc(64 * 128);
		bzero (obj.pix.tree[i], 64 * 128);

		if (!i) {
			obj.pos.tree[0].x = (rand() % 150) + 160;
			obj.pos.tree[0].y = (obj.pos.ground.map[obj.pos.tree[0].x] - 128) + 8;
		} else {
			obj.pos.tree[i].x = (rand() % 320) + 320;
			obj.pos.tree[i].y = 200;	/* evil, ugly hack */
		}

		for (n = 0; n < 8; n++) {
			obj.pos.tree[i].dx[n] = (rand() % (64 - (17 * 2))) + 17;
			obj.pos.tree[i].dy[n] = (rand() % (64 - (17 * 2))) + 17;
		}
	}
}

void init_mouse (void)
{
	obj.pix.mouse[0] = (char *) malloc(32 * 32);
	obj.pix.mouse[1] = (char *) malloc(32 * 32);
	bzero (obj.pix.mouse[0], 32 * 32);
	bzero (obj.pix.mouse[1], 32 * 32);

	obj.pos.mouse.x = 0;
	obj.pos.mouse.y = 0;
	obj.pos.mouse.phase = 0;

	/* xxx: do phase_mouse */

{
	int a;

	for (a = 0; a < 1024; a++) {
		int x, y;

		x = sincos[a ^ 1024] * 13 / 1024 + 13;
		y = sincos[a] * 4 / 1024 + 4;

		x += 3;
		y += 3;

		obj.pix.mouse[0][(y << 5) ^ x] = 1;
	}
}
}

void deinit_sun (void)
{
	free (obj.pix.sun);
}

void deinit_cloud (void)
{
	int i;

	for (i = 0; i < NCLOUD; i++)
		free (obj.pix.cloud[(NCLOUD - i) - 1]);
}

void deinit_groundmap (void)
{
	free (obj.pix.ground);
}

void deinit_tree (void)
{
	int i;

	for (i = 0; i < NTREE; i++)
		free (obj.pix.tree[(NTREE - i) - 1]);
}

void deinit_mouse (void)
{
	free (obj.pix.mouse[1]);
	free (obj.pix.mouse[0]);
}

void move_sun (void)
{
	struct sun_type *s = &obj.pos.sun;

	if (pstate == ST_FADEOUT) {
		if (s->opstate != ST_FADEOUT) {
			s->opstate = ST_FADEOUT;
			s->saved_bx = s->bx;
		}

		if (s->bx > -64)
			s->bx -= (s->saved_bx > (160 - 32)) ? 1 : 2;
	} else if (s->bx > (160 - 32))
		s->bx--;

	if (s->hop) {
		s->x = s->bx + (sincos[s->ang ^ 1024] * (s->hop >> 5) >> 10);
		s->y = (sincos[s->ang] * (s->hop >> 3) >> 10);
		s->ang += 16;
		s->ang &= 1023;

		if (s->bx == (160 - 32)) {
			if ((s->hop >> 3) > 3)
				s->hop--;
		} else if (s->bx < (160 - 32)) {
			if ((s->hop >> 3) < 16)
				s->hop++;
		}
	}
}

void move_cloud (void)
{
	int i;
	static char div = 0;
	struct cloud_type *c = obj.pos.cloud;

	div++;
	div %= NCLOUD;

	for (i = 0; i < NCLOUD; i++) {
		if (div >= i) {
			if (c->x > -64)
				c->x -= (rand() & 1) + 1;
			else
				c->x = 320;
		}
		c++;
	}
}

void move_groundmap (void)
{
	int x, y, d;

	for (x = 0; x < 319; x++)
		obj.pos.ground.map[x] = obj.pos.ground.map[x + 1];

	y = obj.pos.ground.map[319];
	d = (rand() % 3) - 1;
	if ((y + d) > (150 + 32) || (y + d) < 150)
		y -= d;
	else
		y += d;
	obj.pos.ground.map[319] = y;

	obj.pos.ground.shift++;
	obj.pos.ground.shift &= 31;
}

void move_tree (void)
{
	int i;
	struct tree_type *t = obj.pos.tree;

	for (i = 0; i < NTREE; i++) {
		if (t->x > -64) {
			t->x--;
			if (t->x == 319)
				t->y = (obj.pos.ground.map[319] - 128) + 8 + (rand() % 63);
		} else
			t->x = (rand() % 320) + 320;

		t++;
	}
}

void move_mouse (void)
{
	static char div = 0;

	if (++div > 6) {
		div = 0;
		// xxx obj.pos.mouse.phase ^= 1;
	}

	if (keytab[0x48]) {
		if (obj.pos.mouse.y >= -16)
			obj.pos.mouse.y--;
		if (obj.pos.mouse.y >= -16)
			obj.pos.mouse.y--;
	} else if (keytab[0x50]) {
		if (obj.pos.mouse.y < (200 - 16))
			obj.pos.mouse.y++;
		if (obj.pos.mouse.y < (200 - 16))
			obj.pos.mouse.y++;
	}

	if (keytab[0x4B]) {
		if (obj.pos.mouse.x >= -16)
			obj.pos.mouse.x--;
		if (obj.pos.mouse.x >= -16)
			obj.pos.mouse.x--;
	} else if (keytab[0x4D]) {
		if (obj.pos.mouse.x < (320 - 16))
			obj.pos.mouse.x++;
		if (obj.pos.mouse.x < (320 - 16))
			obj.pos.mouse.x++;
	}

	/* xxx: fire */
}

void phase_sun (int phase)
{
	/*
	 * fazy rysowania:
	 *
	 * 1. kko: 64 razy
	 * 2. wypenienie: 64 razy
	 * 3. narysowanie promieni poziomych: 64 razy
	 * 4. narysowanie promieni ukonych: 48 razy
	 * 5. narysowanie oczu: raz
	 * 6. narysowanie umiechu: 2 razy
	 * 7. narysowanie umiechu: 9 razy
	 * 8. narysowanie umiechu: 2 razy
	 * 9. koniec fazy rysowania: raz
	 */

#define	PH_SUN_1 64
#define	PH_SUN_2 (PH_SUN_1 + 64)
#define	PH_SUN_3 (PH_SUN_2 + 64)
#define	PH_SUN_4 (PH_SUN_3 + 48)
#define	PH_SUN_5 (PH_SUN_4 + 1)
#define	PH_SUN_6 (PH_SUN_5 + 2)
#define	PH_SUN_7 (PH_SUN_6 + 9)
#define	PH_SUN_8 (PH_SUN_7 + 2)

	if (phase < PH_SUN_1) {
		int ang, cang = phase << 4;
		int x, y;

		for (ang = 0; ang < 16; ang++) {
			x = sincos[(ang + cang) ^ 1024] * 16 / 1024 + 32;
			y = sincos[ang + cang] * 16 / 1024 + 32;
			obj.pix.sun[(y << 6) + x] = 1;
		}
	} else if (phase < PH_SUN_2) {
		fill_line (obj.pix.sun + ((phase - PH_SUN_1) << 6), 64, 28, 35, 0, 0);
	} else if (phase < PH_SUN_3) {
		int ofs;

		ofs = ((phase - PH_SUN_2) << 6) + 32;
		obj.pix.sun[ofs] = obj.pix.sun[ofs] ? obj.pix.sun[ofs] : 1;
		ofs = (phase - PH_SUN_2) + (32 << 6);
		obj.pix.sun[ofs] = obj.pix.sun[ofs] ? obj.pix.sun[ofs] : 1;
	} else if (phase < PH_SUN_4) {
		int ofs, ph = (phase - PH_SUN_3) + 8;

		ofs = (ph << 6) + ph;
		obj.pix.sun[ofs] = obj.pix.sun[ofs] ? obj.pix.sun[ofs] : 1;
		ofs = (ph << 6) + (63 - ph);
		obj.pix.sun[ofs] = obj.pix.sun[ofs] ? obj.pix.sun[ofs] : 1;
	} else if (phase < PH_SUN_5) {
		obj.pix.sun[(26 << 6) + 26] = 67;
		obj.pix.sun[(26 << 6) + 38] = 67;

		obj.pix.sun[(26 << 6) + 25] = 68;
		obj.pix.sun[(26 << 6) + 27] = 68;
		obj.pix.sun[(25 << 6) + 26] = 68;
		obj.pix.sun[(27 << 6) + 26] = 68;

		obj.pix.sun[(26 << 6) + 37] = 68;
		obj.pix.sun[(26 << 6) + 39] = 68;
		obj.pix.sun[(25 << 6) + 38] = 68;
		obj.pix.sun[(27 << 6) + 38] = 68;
	} else if (phase < PH_SUN_6) {
		if (phase == PH_SUN_5)
			obj.pix.sun[(38 << 6) + 26] = 4;
		else
			obj.pix.sun[(39 << 6) + 27] = 4;
	} else if (phase < PH_SUN_7) {
		obj.pix.sun[(40 << 6) + (phase - PH_SUN_6 + 28)] = 4;
	} else if (phase < PH_SUN_8) {
		if (phase == PH_SUN_7)
			obj.pix.sun[(39 << 6) + 37] = 4;
		else
			obj.pix.sun[(38 << 6) + 38] = 4;
	}
}

void cloud_border (char *ptr, int x, int y)
{
	if (x < 0 || x >= 64 || y < 0 || y >= 32)
		return;

	if (ptr[(y << 6) + x] < 36 || ptr[(y << 6) + x] > 67)
		ptr[(y << 6) + x] = 3;
}

void tree_border (char *ptr, int x, int y)
{
	if (x < 0 || x >= 64 || y < 0 || y >= 64)
		return;

	if (ptr[(y << 6) + x] < 71 || ptr[(y << 6) + x] > 102)
		ptr[(y << 6) + x] = 69;
}

void phase_cloud (int phase)
{
	int i;

	/*
	 * fazy rysowania:
	 *
	 * 1. 8 elips: 64
	 * 2. wypenienie: 32
	 * 3. obrysowanie: 32
	 * 4. koniec rysowania
	 */

#define	PH_C_1 64
#define	PH_C_2 (PH_C_1 + 32)
#define	PH_C_3 (PH_C_2 + 32)

	for (i = 0; i < NCLOUD; i++)
		if (phase < PH_C_1) {
			int cang = phase << 4;
			int ang, x, y;
			int n;

			for (n = 0; n < 8; n++)
				for (ang = 0; ang < 16; ang++) {
					x = sincos[(ang + cang) ^ 1024] * 16 / 1024 + obj.pos.cloud[i].dx[n];
					y = sincos[ang + cang] * 8 / 1024 + obj.pos.cloud[i].dy[n];
					obj.pix.cloud[i][(y << 6) + x] = gcolor(36, 67);
				}
		} else if (phase < PH_C_2) {
			fill_line (obj.pix.cloud[i] + ((phase - PH_C_1) << 6), 64, 36, 67, 1, 0);
		} else if (phase < PH_C_3) {
			int x;
			int y = phase - PH_C_2;
			char *ptr = obj.pix.cloud[i];

			for (x = 0; x < 64; x++) {
				if (ptr[(y << 6) + x] < 36 || ptr[(y << 6) + x] > 67)
					continue;

				cloud_border (ptr, x - 1, y - 1);
				cloud_border (ptr, x, y - 1);
				cloud_border (ptr, x + 1, y - 1);
				cloud_border (ptr, x - 1, y);
				cloud_border (ptr, x, y);
				cloud_border (ptr, x + 1, y);
				cloud_border (ptr, x - 1, y + 1);
				cloud_border (ptr, x, y + 1);
				cloud_border (ptr, x + 1, y + 1);
			}
		}
}

void phase_groundmap (int phase)
{
	/* null */
}

void phase_tree (int phase)
{
	int i;
	static char c[8], cdir[8];

	/*
	 * fazy rysowania:
	 *
	 * 1. kij (zapomniaem jak to si nazywa ;P): 128 - 32
	 * 2. osiem kek na pozycjach z dx, dy: 128
	 * 3. wypenienie tych obszarw: 64
	 * 4. narysowanie obramowania: 64
	 * 5. koniec rysowania
	 */

#define	PH_T_1 (128 - 32)
#define	PH_T_2 (PH_T_1 + 128)
#define	PH_T_3 (PH_T_2 + 64)
#define	PH_T_4 (PH_T_3 + 64)

	if (!phase)
		for (i = 0; i < NTREE; i++) {
			c[i] = 4;
			cdir[i] = 1;
		}

	for (i = 0; i < NTREE; i++)
		if (phase < PH_T_1) {
			int x, y = ((127 - 32) - phase) + 32;

			obj.pix.tree[i][(y << 6) + 28] = 4;
			obj.pix.tree[i][(y << 6) + 35] = 4;

			for (x = 29; x < 35; x++) {
				obj.pix.tree[i][(y << 6) + x] = c[i];
				if (rand() & 1)
					if (cdir[i])
						if (c[i] < 35)
							c[i]++;
						else
							cdir[i] ^= 1;
					else
						if (c[i] > 4)
							c[i]--;
						else
							cdir[i] ^= 1;
			}
		} else if (phase < PH_T_2) {
			int x, y, n;
			int ang, cang = (phase - PH_T_1) << 3;

			for (n = 0; n < 8; n++)
				for (ang = 0; ang < 8; ang++) {
					x = sincos[(ang + cang) ^ 1024] * 16 / 1024 + obj.pos.tree[i].dx[n];
					y = sincos[ang + cang] * 16 / 1024 + obj.pos.tree[i].dy[n];
					obj.pix.tree[i][(y << 6) + x] = 71 + (rand() & 31);
				}
		} else if (phase < PH_T_3) {
			fill_line (obj.pix.tree[i] + ((phase - PH_T_2) << 6), 64, 71, 102, 1, 1);
		} else if (phase < PH_T_4) {
			int x;
			int y = phase - PH_T_3;
			char *ptr = obj.pix.tree[i];

			for (x = 0; x < 64; x++) {
				if (ptr[(y << 6) + x] < 71 || ptr[(y << 6) + x] > 102)
					continue;

				tree_border (ptr, x - 1, y - 1);
				tree_border (ptr, x, y - 1);
				tree_border (ptr, x + 1, y - 1);
				tree_border (ptr, x - 1, y);
				tree_border (ptr, x, y);
				tree_border (ptr, x + 1, y);
				tree_border (ptr, x - 1, y + 1);
				tree_border (ptr, x, y + 1);
				tree_border (ptr, x + 1, y + 1);
			}
		}
}

void phase_mouse (int phase)
{
/*
	int i;
	char ch;

	for (i = 0; i < 32 * 32; i++) {
		ch = rand();
		if (ch & 0x80)
			ch = 0;
		obj.pix.mouse[0][i] = ch;
		ch = rand();
		if (ch & 0x80)
			ch = 0;
		obj.pix.mouse[1][i] = ch;
	}
*/

	// xxx
}

void draw_sun (struct sun_type *p)
{
	draw_picture (obj.pix.sun, 64, 64, p->x, p->y);
}

void draw_cloud (struct cloud_type *p)
{
	int i;

	for (i = 0; i < NCLOUD; i++) {
		draw_picture (obj.pix.cloud[i], 64, 32, p->x, p->y);
		p++;
	}
}

char ground_pix (int x, int y)
{
	if (x & 16)
		x = 15 - (x & 15);
	else
		x &= 15;

	if (y & 16)
		y = 15 - (y & 15);
	else
		y &= 15;

	return obj.pix.ground[(y << 4) ^ x];
}

void draw_groundmap (struct ground_type *p)
{
	int x, y, yp;
	char *m = p->map;

	for (x = 0; x < 320; x++) {
		yp = *m++;
		vbuf[(yp << 8) + (yp << 6) + x] = 70;
		for (y = yp + 1; y < 200; y++)
			vbuf[(y << 8) + (y << 6) + x] = ground_pix(x + p->shift, y);
	}
}

void draw_tree (struct tree_type *p)
{
	int i;

	for (i = 0; i < NTREE; i++) {
		draw_picture (obj.pix.tree[i], 64, 128, p->x, p->y);
		p++;
	}
}

void draw_mouse (struct mouse_type *m)
{
	draw_picture (obj.pix.mouse[m->phase], 32, 32, m->x, m->y);
}

void init_objects (void)
{
	srand (time(NULL));

	init_sun();
	init_cloud();
	init_groundmap();
	init_tree();
	init_mouse();

	obj.pos.phase = 0;
}

void deinit_objects (void)
{
	deinit_mouse();
	deinit_tree();
	deinit_groundmap();
	deinit_cloud();
	deinit_sun();
}

void move_objects (void)
{
	static char div = 0;

	if (!can_move)
		return;

	move_sun();
	move_cloud();
	if (++div > 1) {
		div = 0;
		move_groundmap();
		move_tree();
	}
	move_mouse();

	if (obj.pos.phase < MAX_PHASE)
		obj.pos.phase++;
}

void draw_objects (void)
{
	struct pos_type p;
	static int ophase = 0;

	can_move = 0;
	memcpy (&p, &obj.pos, sizeof(struct pos_type));
	can_move = 1;

	while (ophase < p.phase) {
		phase_sun (ophase);
		phase_cloud (ophase);
		phase_groundmap (ophase);
		phase_tree (ophase);
		phase_mouse (ophase);

		ophase++;
	}

	draw_sun (&p.sun);
	draw_cloud (&p.cloud);
	draw_mouse (&p.mouse);
	draw_tree (&p.tree);
	draw_groundmap (&p.ground);
}
